import random
import sys
from time import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from fisher import pvalue
from scipy.stats import chi2_contingency

from unpast.core.feature_clustering import run_Louvain
from unpast.core.sample_clustering import cluster_samples, update_bicluster_data
from unpast.misc.eval import find_best_matching_biclusters


def make_consensus_biclusters(
    biclusters_list,
    exprs,
    similarity="samples",  # can be 'both','genes','samples'
    p=0.05,  # bicluster overlap significance threshold
    min_similarity=1 / 3,  # minimal Jaccard similarity
    max_similarity=0.9,  # maximal Jaccard similarity
    frac_runs=1 / 3,
    min_n_genes=2,
    min_n_samples=5,
    min_n_times_detected=2,
    modularity_measure="newman",
    method="kmeans",  # sample clustering method
    seed=42,  # fixed by default, can be set to -1 to use random seed
    plot=False,
    figsize=(7, 7),
    labels=False,
    colorbar_off=True,
    verbose=True,
):
    # considers only significantly overlapping and best matching bicluster pairs
    # list of biclusters from several runs
    n_runs = len(biclusters_list)
    if n_runs < min_n_times_detected:
        print(
            "The number of biclusterins results should be not less than min_n_times_detected=%s"
            % (min_n_times_detected),
            file=sys.stderr,
        )
        return

    t0 = time()
    N = exprs.shape[0]

    if seed == -1:
        seed = random.randint(0, 1000000)
        print("Seed for sample clustering: %s" % (seed), file=sys.stderr)

    for i in range(n_runs):
        biclusters_list[i]["run"] = i
    biclusters = pd.concat(biclusters_list)
    bic_ids = biclusters.index.values
    biclusters["detected_n_times"] = 1

    n_bics = biclusters.shape[0]
    J_heatmap = pd.DataFrame(np.zeros((n_bics, n_bics)), index=bic_ids, columns=bic_ids)

    # add only best matches to Jaccard similarity matrix
    avg_J_sim = {}
    for i in range(n_runs):
        # bicluster is always the best match of itself,
        # similarity matrix block for output of a biclustering method w. itself is an identity matrix
        bics1 = biclusters.loc[biclusters["run"] == i, :]
        J_heatmap.loc[bics1.index.values, bics1.index.values] = np.identity(
            bics1.shape[0]
        )
        avg_J_sim[i] = {i: 1}
        for j in range(n_runs):
            if i != j:
                bics2 = biclusters.loc[biclusters["run"] == j, :]
                # find best matches between bics1 and bics2
                bm = find_best_matching_biclusters(
                    bics1,
                    bics2,
                    exprs.shape,
                    by=similarity,
                    min_g=min_n_genes,
                    adj_pval_thr=p,
                )
                bm = bm.dropna()
                if bm.shape[0] > 0:
                    avg_J_sim[i][j] = np.mean(bm["J"])
                    df = bm.loc[:, ["bm_id", "J"]]
                    for row in df.iterrows():
                        J_heatmap.loc[row[0], row[1]["bm_id"]] += row[1]["J"] / 2
                        J_heatmap.loc[row[1]["bm_id"], row[0]] += row[1]["J"] / 2

    # if all biclusters are exactly the same
    if J_heatmap.min().min() == 1:
        # return the first bicluster
        consensus_biclusters = biclusters.iloc[[0], :].copy()
        consensus_biclusters.index = [0]
        consensus_biclusters.loc[0, "detected_n_times"] = biclusters.shape[0]
        print("all biclusters are exactly the same", file=sys.stderr)
        return consensus_biclusters

    # plot bicluster similarity heatmaps
    if plot:
        import seaborn as sns

        avg_J_sim = pd.DataFrame.from_dict(avg_J_sim)
        avg_J_sim = avg_J_sim.fillna(0)
        if avg_J_sim.shape[0] > 0:
            g = sns.clustermap(
                avg_J_sim,
                linewidths=0,
                vmin=0,
                vmax=1,
                figsize=(3, 3),
                center=0,
                annot=True,
            )
            g.ax_cbar.set_visible(False)
            plt.show()

        labels = True
        if len(bic_ids) > 20:
            labels = False
        g = sns.clustermap(
            J_heatmap,
            yticklabels=labels,
            xticklabels=labels,
            linewidths=0,
            vmin=0,
            vmax=1,
            figsize=figsize,
            center=0,
            annot=labels,
        )
        g.ax_row_dendrogram.set_visible(False)
        g.ax_col_dendrogram.set_visible(False)
        if colorbar_off:
            g.cax.set_visible(False)
        plt.show()

    t1 = time()
    if verbose:
        print("%s s for similarity matrix" % round(t1 - t0))

    # cluster biclusters by similarity
    matched, not_matched, cutoff = run_Louvain(
        J_heatmap,
        similarity_cutoffs=np.arange(min_similarity, max_similarity, 0.05),
        m=False,
        plot=plot,
        modularity_measure=modularity_measure,
    )
    t2 = time()

    # make consensus biclusters
    # for each group of matched biclusters, select consensus gene set
    # keep genes occuring at least 'min_n_times_detected' times
    min_n_times_detected = max(min_n_times_detected, int(frac_runs * n_runs))
    if verbose:
        print(
            "keep genes included in at least %s merged biclusters"
            % round(min_n_times_detected)
        )

    consensus_biclusters = []
    # for each group of matched biclusters
    for i in range(len(matched)):
        gsets = biclusters.loc[matched[i], "genes"].values
        bic_ids = biclusters.loc[matched[i], :].index.values
        detected_n_times = biclusters.loc[matched[i], "detected_n_times"].sum()
        # count gene occurencies
        gene_occurencies = {}
        for gene in set().union(*gsets):
            gene_occurencies[gene] = 0
            for gset in gsets:
                if gene in gset:
                    gene_occurencies[gene] += 1

        gene_occurencies = pd.Series(gene_occurencies).sort_values()
        passed_genes = sorted(
            gene_occurencies[
                gene_occurencies >= min(n_runs, len(matched)) * frac_runs
            ].index.values
        )
        not_passed_genes = sorted(
            gene_occurencies[
                gene_occurencies < min(n_runs, len(matched)) * frac_runs
            ].index.values
        )

        if len(passed_genes) < min_n_genes:
            # move all biclusters to not matched
            not_matched += list(matched[i])
        else:
            # cluster samples again in a subspace of a new gene set
            bicluster = cluster_samples(
                exprs.loc[passed_genes, :].T,
                min_n_samples=min_n_samples,
                seed=seed,
                method=method,
            )
            # if bicluster is not empty, add it to consenesus
            if "sample_indexes" in bicluster.keys():
                bicluster["genes"] = set(passed_genes)
                bicluster["n_genes"] = len(bicluster["genes"])
                bicluster = update_bicluster_data(bicluster, exprs)
                bicluster["detected_n_times"] = detected_n_times
                bicluster["ids"] = set(
                    bic_ids
                )  # ids of biclusters merged to the consensus biclusters
                consensus_biclusters.append(bicluster)
    consensus_biclusters = pd.DataFrame.from_records(consensus_biclusters)

    # add not matched
    not_changed_biclusters = biclusters.loc[not_matched, :]
    # not_changed_biclusters["detected_n_times"] = 1
    consensus_biclusters = pd.concat([consensus_biclusters, not_changed_biclusters])

    # add direction
    consensus_biclusters["direction"] = "BOTH"
    consensus_biclusters.loc[
        consensus_biclusters["n_genes"] == consensus_biclusters["genes_up"].apply(len),
        "direction",
    ] = "UP"
    consensus_biclusters.loc[
        consensus_biclusters["n_genes"]
        == consensus_biclusters["genes_down"].apply(len),
        "direction",
    ] = "DOWN"

    # sort
    col_order = [
        "SNR",
        "n_genes",
        "n_samples",
        "genes",
        "samples",
        "genes_up",
        "genes_down",
        "gene_indexes",
        "sample_indexes",
        "direction",
        "detected_n_times",
        "ids",
    ]
    consensus_biclusters = consensus_biclusters.sort_values(
        by=["SNR", "n_samples"], ascending=[False, True]
    )
    consensus_biclusters = consensus_biclusters.loc[:, col_order]
    consensus_biclusters = consensus_biclusters.loc[
        consensus_biclusters["n_samples"] >= min_n_samples, :
    ]
    if verbose:
        print("all consensus biclusters:", consensus_biclusters.shape[0])

    consensus_biclusters = consensus_biclusters.loc[
        consensus_biclusters["detected_n_times"] >= min_n_times_detected, :
    ]

    if verbose:
        print(
            "detected %s+ times:%s"
            % (min_n_times_detected, consensus_biclusters.shape[0])
        )

    consensus_biclusters.index = range(consensus_biclusters.shape[0])

    if verbose:
        print(
            round(time() - t2),
            "s for making consensus biclusters from consensus gene sets",
        )

    return consensus_biclusters


def calc_signif_bicluster_similarities(
    biclusters_dict,
    exprs,
    similarity="both",
    adj_pval_thr=0.05,
    plot=True,
    figsize=(10, 10),
    labels=False,
    colorbar_off=True,
):
    """WILL REPLACE find_best_matching_biclusters(), currently is not used.
    Calculates Jaccard similarities of significanly overlapping bicluster paires based on gene and/or sample overlaps, and optionally visualize the results.

        Parameters
        ----------
        biclusters_dict : dict
            A dictionary of bicluster DataFrames. A bicluster DataFrame where each row represents a bicluster, and columns contain sets of 'genes' and 'samples' associated with the bicluster.
            Keys of the dictionary may correspond to run names.

        exprs : pd.DataFrame
            Gene expression data with rows corresponding to genes and columns to samples. Used for determining overlaps between biclusters.

        similarity : str, optional, default="both"
            The type of similarity to compute:
            - "genes": only compute similarities based on gene overlaps.
            - "samples": only compute similarities based on sample overlaps.
            - "both": compute similarities based on both gene and sample overlaps.

        adj_pval_thr : float, optional, default=0.05
            Adjusted p-value threshold for significance testing. Similarity values will be stored only for bicluster pairs which significantly overlap. Statistical significance of the overlaps is tested using the chi-squared test.

        plot : bool, optional, default=True
            If True, generate a heatmap visualization of the similarity matrix.

        figsize : tuple, optional, default=(10, 10)
            Size of the heatmap plot, used when `plot=True`.

        labels : bool, optional, default=False
            If True, display labels for the biclusters on the axes of the heatmap.

        colorbar_off : bool, optional, default=True
            If True, hide the colorbar in the heatmap visualization.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing pairwise Jaccard similarity values between biclusters, adjusted for statistical significance.
    """

    if similarity not in ["genes", "samples", "both"]:
        print(
            "Similarity must be 'genes','samples','both'. Set to 'both'.",
            file=sys.stderr,
        )
        similarity = "both"

    J_heatmap = {}
    s = set(exprs.columns.values)
    g = set(exprs.index.values)
    N_bics = len(biclusters_dict.keys())
    ### plot pairwise comparisons:
    for bic_n1 in biclusters_dict.keys():
        b1 = biclusters_dict[bic_n1]
        g1 = b1["genes"]
        s1 = b1["samples"]
        J_heatmap[bic_n1] = {}
        for bic_n2 in biclusters_dict.keys():
            b2 = biclusters_dict[bic_n2]
            g2 = b2["genes"]
            s2 = b2["samples"]
            g_overlap = g1.intersection(g2)
            g_union = g1.union(g2)
            s_overlap = s1.intersection(s2)
            s_union = s1.union(s2)
            J_g = len(g_overlap) / len(g_union)
            J_s = len(s_overlap) / len(s_union)
            J_heatmap[bic_n1][bic_n2] = 0

            # significance of gene overlap
            if similarity != "samples":
                if len(g_overlap) == 0:
                    p_g = 1
                else:
                    chi2, p_g, dof, expected = chi2_contingency(
                        [
                            [len(g_overlap), len(g1.difference(g2))],
                            [len(g2.difference(g1)), len(g.difference(g1 | g2))],
                        ]
                    )

            # significance of sample overlap
            if similarity != "genes":
                # skip if similarity==both and gene overlap is empty
                if similarity == "both" and len(g) == 0:
                    p_s = 1
                else:
                    chi2, p_s, dof, expected = chi2_contingency(
                        [
                            [len(s_overlap), len(s1.difference(s2))],
                            [len(s2.difference(s1)), len(s.difference(s1 | s2))],
                        ]
                    )

            if similarity == "genes":
                if p_g * (N_bics - 1) * N_bics / 2 < adj_pval_thr:
                    J_heatmap[bic_n1][bic_n2] = J_g
            elif similarity == "samples":
                if p_s * (N_bics - 1) * N_bics / 2 < adj_pval_thr:
                    J_heatmap[bic_n1][bic_n2] = J_s
            elif similarity == "both":  # both genes and samples considered
                # consider significant overlaps in g and s and save max. J
                if (
                    p_g * (N_bics - 1) * N_bics / 2 < adj_pval_thr
                    and len(s_overlap) > 0
                ) or (
                    p_s * (N_bics - 1) * N_bics / 2 < adj_pval_thr
                    and len(g_overlap) > 0
                ):
                    J_heatmap[bic_n1][bic_n2] = max(J_s, J_g)

    J_heatmap = pd.DataFrame.from_dict(J_heatmap)

    if plot:
        import seaborn as sns

        g = sns.clustermap(
            J_heatmap,
            yticklabels=labels,
            xticklabels=labels,
            linewidths=0,
            vmin=0,
            vmax=1,
            figsize=figsize,
            center=0,
            annot=labels,
        )
        g.ax_row_dendrogram.set_visible(False)
        g.ax_col_dendrogram.set_visible(False)
        if colorbar_off:
            g.cax.set_visible(False)
        plt.show()

    return J_heatmap



"""Tests for run_unpast, and hence all the core code. Usage: python -m pytest test/test_run_unpast.py"""

import os
import sys

import pandas as pd
import pytest

TEST_DIR = os.path.dirname(__file__)
RESULTS_DIR = os.path.join(TEST_DIR, "results")
if not os.access(RESULTS_DIR, os.W_OK):
    # repo dir is currently read-only during the testing stage in github-action
    RESULTS_DIR = "/tmp/unpast/results"
REFERENCE_OUTPUT_DIR = os.path.join(TEST_DIR, "test_reference_output")

from unpast.run_unpast import unpast

### Helper functions ###


def run_unpast_on_file(filename, basename, *args, **kwargs):
    out_dir = os.path.join(RESULTS_DIR, f"runs/run_{basename}")
    unpast(
        os.path.join(TEST_DIR, filename),
        out_dir=out_dir,
        verbose=True,  # use pytest -s ... to see the output
        *args,
        **kwargs,
    )
    return parse_answer(out_dir)


def parse_answer(answer_dir, startswith=""):
    files = os.listdir(answer_dir)
    answer_files = [
        f for f in files if f.startswith(startswith) and f.endswith("biclusters.tsv")
    ]
    assert len(answer_files) == 1, f"There are {len(answer_files)} files instead of 1"
    return pd.read_csv(os.path.join(answer_dir, answer_files[0]), sep="\t", comment="#")


def parse_to_features_samples_ids(answer):
    def to_set_of_nums(s):
        return set(map(int, s.strip().split()))

    return (
        to_set_of_nums(answer["gene_indexes"]),
        to_set_of_nums(answer["sample_indexes"]),
    )


### Tests ###


@pytest.mark.slow
def test_smoke():
    """Smoke test - check that the program runs on some input without failure."""
    run_unpast_on_file(
        filename="test_input/synthetic_clear_biclusters.tsv",
        basename="test_smoke",
        clust_method="WGCNA",
    )


def test_simple():
    """Check that clear biclusters are found."""
    res = run_unpast_on_file(
        filename="test_input/synthetic_small_example.tsv",
        basename="test_simple",
        min_n_samples=2,
        clust_method="Louvain",
    )
    assert len(res) == 1, "Too many clusters found"
    features, samples = parse_to_features_samples_ids(res.iloc[0])
    assert features == {0, 1}
    assert samples == {0, 1}


@pytest.mark.slow
def test_clear_biclusters():
    """Check that clear biclusters are found."""
    res = run_unpast_on_file(
        filename="test_input/synthetic_clear_biclusters.tsv",
        basename="test_clear_biclusters",
        clust_method="WGCNA",
    )

    found_correct_bicluster = False
    for _, row in res.iterrows():
        features, samples = parse_to_features_samples_ids(row)
        if features == set(range(1, 22, 2)) and samples == set(range(1, 22, 2)):
            found_correct_bicluster = True

    assert found_correct_bicluster


@pytest.mark.slow
def test_reproducible_wgcna():
    """Check that the same data is found on a complicated input with no clear answer."""
    res = run_unpast_on_file(
        filename="test_input/synthetic_noise.tsv",
        basename="test_reproducible_wgcna",
        clust_method="WGCNA",
        dch=0.999,  # no biclusters found in noise otherwise
    )
    reference = parse_answer(
        answer_dir=REFERENCE_OUTPUT_DIR,
        startswith="test_reproducible_wgcna",
    )
    assert res.equals(reference), "The results are not reproducible"


def test_reproducible_louvain():
    """Check that the same data is found on a complicated input with no clear answer."""
    res = run_unpast_on_file(
        filename="test_input/synthetic_noise.tsv",
        basename="test_reproducible_louvain",
        clust_method="Louvain",
        similarity_cutoffs=[0.5],  # no biclusters found in noise otherwise
    )
    reference = parse_answer(
        answer_dir=REFERENCE_OUTPUT_DIR,
        startswith="test_reproducible_louvain",
    )
    assert res.equals(reference), "The results are not reproducible"



import os
import tempfile

import pandas as pd
import pytest

from unpast.utils import io


# Helper to create a minimal bicluster DataFrame
def make_bicluster_df():
    return pd.DataFrame(
        {
            "genes": [set(["g1", "g2"])],
            "genes_up": [set(["g1"])],
            "genes_down": [set(["g2"])],
            "samples": [set(["s1", "s2"])],
            "gene_indexes": [set([0, 1])],
            "sample_indexes": [set([0, 1])],
        }
    )


def test_read_write_bic_table(tmp_path):
    df = make_bicluster_df()
    file_path = tmp_path / "biclusters.tsv"
    io.write_bic_table(df, str(file_path))
    result = io.read_bic_table(str(file_path))
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 1
    assert result.iloc[0]["genes"] == set(["g1", "g2"])
    assert result.iloc[0]["genes_up"] == set(["g1"])
    assert result.iloc[0]["genes_down"] == set(["g2"])
    assert result.iloc[0]["samples"] == set(["s1", "s2"])
    assert result.iloc[0]["gene_indexes"] == set([0, 1])
    assert result.iloc[0]["sample_indexes"] == set([0, 1])


def test_read_bic_table_missing_file():
    with pytest.raises(FileNotFoundError):
        io.read_bic_table("nonexistent_file.tsv")


def test_read_bic_table_empty(tmp_path):
    empty_df = make_bicluster_df().iloc[0:0]
    file_path = tmp_path / "empty.tsv"

    io.write_bic_table(empty_df, str(file_path))
    result = io.read_bic_table(str(file_path))

    assert isinstance(result, pd.DataFrame)
    assert result.empty


def test_write_bic_table_metadata(tmp_path):
    df = make_bicluster_df()
    file_path = tmp_path / "biclusters_meta.tsv"
    io.write_bic_table(
        df,
        str(file_path),
        add_metadata=True,
        seed=123,
        pval=0.05,
        min_n_samples=2,
        bin_method="kmeans",
        clust_method="Louvain",
        directions=["UP", "DOWN"],
        similarity_cutoff=0.9,
        m=0.33,
        merge=True,
    )
    with open(file_path) as f:
        first_line = f.readline()
    assert first_line.startswith("#")


def test_project_paths(tmp_path):
    paths = io.ProjectPaths(str(tmp_path))
    assert os.path.exists(paths.root)
    assert paths.res.endswith(".tsv")

    root = paths.get_root_dir()
    new_paths = io.ProjectPaths(root)
    assert new_paths.get_root_dir() == root

    assert not os.path.exists(paths.bin_dir)
    paths.create_binarization_paths()
    assert os.path.exists(paths.bin_dir)


def test_write_and_read_args(tmp_path):
    args = {
        "float_value": 0.1,
        "str_value": "test",
        "ing_value": 42,
        "bool_value": True,
        "list_value": [1, 2, 3],
    }
    file_path = tmp_path / "args.tsv"
    io.write_args(args, str(file_path))
    loaded_args = io.read_args(str(file_path))
    # All values are read as strings from the file
    expected_args = {k: str(v) for k, v in args.items()}
    assert loaded_args == expected_args



"""TODO: check and remove this file if not used anymore"""

from .logs import get_logger

logger = get_logger(__name__)

TRY_USE_NUMBA = True


def jit_if_available(func):
    """Decorator that applies numba JIT compilation if available, otherwise returns unmodified function.

    Args:
        func: function to potentially JIT compile

    Returns:
        function: JIT-compiled function if numba available, otherwise original function
    """

    # default "do nothing" decorator with the numba-like interface
    def decorated(*args, **kwargs):
        return func(*args, **kwargs)

    return decorated


if TRY_USE_NUMBA:
    try:
        from numba import jit  # as jit_if_available

        jit_if_available = jit()
    except:
        logger.info(
            "Numba is not available. Install numba for a bit faster calculations"
        )



import matplotlib.pyplot as plt


def plot_binarized_feature_impl(
    feature_name, down_group, up_group, colors, hist_range, snr
):
    """Plot histogram of binarized feature values showing up and down groups.

    Args:
        feature_name (str): name of the feature being plotted
        down_group (array): values in the down-regulated group
        up_group (array): values in the up-regulated group
        colors (tuple): colors for (down, up) groups
        hist_range (tuple): range for histogram bins
        snr (float): signal-to-noise ratio value

    Returns:
        None: displays plot
    """
    down_color, up_color = colors
    n_bins = int(max(20, (len(down_group) + len(up_group)) / 10))
    n_bins = min(n_bins, 200)
    fig, ax = plt.subplots()
    tmp = ax.hist(
        down_group, bins=n_bins, alpha=0.5, color=down_color, range=hist_range
    )
    tmp = ax.hist(up_group, bins=n_bins, alpha=0.5, color=up_color, range=hist_range)
    # tmp = plt.title("{}:    SNR={:.2f},    neg={}, pos={}".format(feature_name,snr,len(down_group),len(up_group)))
    n_samples = min(len(down_group), len(up_group))
    # tmp = ax.set_title("SNR={:.2f},   n_samples={}".format(snr,n_samples))
    ax.text(
        0.05,
        0.95,
        feature_name,
        ha="left",
        va="top",
        transform=ax.transAxes,
        fontsize=24,
    )
    ax.text(
        0.95,
        0.95,
        "SNR=" + str(round(snr, 2)) + "\nn_samples=" + str(n_samples),
        ha="right",
        va="top",
        transform=ax.transAxes,
        fontsize=14,
    )
    # tmp = plt.savefig("figs_binarization/"+feature_name+".hist.svg", bbox_inches='tight', transparent=True)
    plt.show()


def plot_binarized_feature(row, direction, pos_mask, neg_mask, snr):
    gene = row.name
    row_values = row.values
    hist_range = row_values.min(), row_values.max()

    # set colors to two sample groups
    # red - overexpression
    # blue - under-expression
    # grey - background (group size > 1/2 of all samples)
    colors = ["grey", "grey"]

    if direction == "UP":
        colors[1] = "red"
    elif direction == "DOWN":
        colors[0] = "blue"
    # elif BOTH ...
    else:
        raise ValueError(f"direction = {direction}")

    # plotting
    plot_binarized_feature_impl(
        gene, row[pos_mask], row[neg_mask], colors, hist_range, snr
    )


def plot_binarization_results(stats, size_snr_trend, sizes, pval):
    """Plot binarization results showing SNR vs sample size with significance threshold.

    Args:
        stats (DataFrame): statistics DataFrame with columns 'SNR', 'SNR_threshold', 'size'
        size_snr_trend (function): function that computes SNR threshold for given sample sizes
        sizes (array): array of sample sizes for plotting the threshold line
        pval (float): p-value threshold used for significance testing

    Returns:
        None: displays plot
    """
    fig, ax = plt.subplots(figsize=(10, 4.5))

    # plot binarization results for real genes
    passed = stats.loc[stats["SNR"] > stats["SNR_threshold"], :]
    failed = stats.loc[stats["SNR"] <= stats["SNR_threshold"], :]
    ax.scatter(
        failed["size"], failed["SNR"], alpha=1, color="black", label="not passed"
    )
    for i, txt in enumerate(failed["size"].index.values):
        ax.annotate(txt, (failed["size"][i], failed["SNR"][i] + 0.1), fontsize=18)
    ax.scatter(passed["size"], passed["SNR"], alpha=1, color="red", label="passed")
    for i, txt in enumerate(passed["size"].index.values):
        ax.annotate(txt, (passed["size"][i], passed["SNR"][i] + 0.1), fontsize=18)

    # plot cutoff
    ax.plot(
        sizes,
        [size_snr_trend(x) for x in sizes],
        color="darkred",
        lw=2,
        ls="--",
        label="e.pval<" + str(pval),
    )
    ax.legend(("not passed", "passed", "e.pval<" + str(pval)), fontsize=18)
    ax.set_xlabel("n_samples", fontsize=18)
    ax.yaxis.tick_right()
    ax.set_ylabel("SNR", fontsize=18)
    ax.set_ylim(0, 4)
    # plt.savefig("figs_binarization/dist.svg", bbox_inches='tight', transparent=True)
    plt.show()



import os
from datetime import datetime
from pathlib import Path
from typing import Optional

import pandas as pd

from .logs import get_logger

logger = get_logger(__name__)
DEFAULT_SAVE_DIR = "."


def read_bic_table(file_name: str, parse_metadata: bool = False) -> pd.DataFrame:
    """
    Reads a bicluster table from a tab-separated file and processes the data into a pandas DataFrame.
    Optionally, it can also parses metadata from the first line of the file if available.

    Args:
        file_name (str): The path to the tab-separated bicluster file.
        parse_metadata (bool, optional): If True, parses metadata from the first line of the file.
                                         The metadata line must start with a "#" and contain key-value
                                         pairs separated by "; ". Defaults to False.

    Returns:
        pandas.DataFrame: A DataFrame containing the bicluster data with columns such as:
                          - "genes": set of all genes in the bicluster.
                          - "genes_up": set of upregulated genes.
                          - "genes_down": set of downregulated genes.
                          - "samples": set of all samples in the bicluster.
                          - "gene_indexes": set of gene index numbers (as integers).
                          - "sample_indexes": set of sample index numbers (as integers).

        If `parse_metadata` is True and metadata exists, returns a tuple:
        (pandas.DataFrame, dict): The bicluster DataFrame and a dictionary of metadata.

    Notes:
        - If the file does not exist returns None, if the file exists but is empty, an empty DataFrame is returned.
        - Missing values in "genes_up" and "genes_down" columns are filled with empty strings.
        - The "genes", "genes_up", "genes_down", "samples", "gene_indexes", and "sample_indexes"
          fields are processed as sets, splitting the original space-separated string values.

    Example:
        biclusters, metadata = read_bic_table('biclusters.tsv', parse_metadata=True)
    """

    biclusters = pd.read_csv(
        file_name,
        sep="\t",
        index_col=0,
        comment="#",
        dtype={
            "sample_indexes": str,
            "gene_indexes": str,
        },
    )
    if len(biclusters) == 0:
        return pd.DataFrame()
    else:
        biclusters["genes"] = biclusters["genes"].apply(
            lambda x: set([g for g in x.split(" ") if not g == ""])
        )
        biclusters["genes_up"] = (
            biclusters["genes_up"]
            .fillna("")
            .apply(lambda x: set([g for g in x.split(" ") if not g == ""]))
        )
        biclusters["genes_down"] = (
            biclusters["genes_down"]
            .fillna("")
            .apply(lambda x: set([g for g in x.split(" ") if not g == ""]))
        )
        biclusters["samples"] = biclusters["samples"].apply(lambda x: set(x.split(" ")))
        if "gene_indexes" in biclusters.columns:
            biclusters["gene_indexes"] = biclusters["gene_indexes"].apply(
                lambda x: set(map(int, set(x.split(" "))))
            )
        if "sample_indexes" in biclusters.columns:
            biclusters["sample_indexes"] = biclusters["sample_indexes"].apply(
                lambda x: set(map(int, set(x.split(" "))))
            )

    if parse_metadata:
        f = open(file_name, "r")
        metadata = f.readline()
        f.close()
        if metadata.startswith("#"):
            metadata = metadata.replace("#", "").rstrip()
            metadata = metadata.split("; ")
            metadata = dict([x.split("=") for x in metadata])
            return biclusters, metadata
    return biclusters


def write_bic_table(
    biclusters: pd.DataFrame,
    results_file_name: str,
    to_str: bool = True,
    add_metadata: bool = False,
    seed: int = None,
    min_n_samples: int = None,
    bin_method: str = None,
    clust_method: str = None,
    pval: float = None,
    directions: list = [],
    similarity_cutoff: float = None,
    ds: float = None,
    dch: float = None,
    m: float = None,
    max_power: int = None,
    precluster: str = None,
    merge: bool = None,
) -> None:
    """
    Writes a bicluster table to a tab-separated file and optionally adds metadata to the file.

    Args:
        biclusters (pd.DataFrame): A DataFrame containing biclusters.
        results_file_name (str): The path where the results should be written.


        to_str (bool, optional): If True, converts sets of values (genes, samples, etc.)
                                 into space-separated strings before writing. Defaults to True.
        add_metadata (bool, optional): Defaults to False. If True, writes metadata to the file's first line.
                                       For example, if `add_metadata` is True, metadata is written to the file
                                        in a format like "#seed=123; pval=0.05; min_n_samples=5; ...".
        seed (int, optional): Seed used for generating biclusters, added to metadata if provided.
        min_n_samples (int, optional): Minimum number of samples, included in metadata.
        bin_method (str, optional): Binning method used, included in metadata.
        clust_method (str, optional): Clustering method used (e.g., "Louvain" or "WGCNA"),
                                      included in metadata.
        pval (float, optional): P-value threshold used for significance, included in metadata.
        directions (list, optional): Directions used in biclustering (e.g., "up", "down"),
                                     included in metadata.
        similarity_cutoff (float, optional): Similarity cutoff used in the Louvain method,
                                             included in metadata.
        ds (float, optional): Soft-thresholding power used in WGCNA, included in metadata.
        dch (float, optional): Dynamic cut height used in WGCNA, included in metadata.
        m (float, optional): Modularity value used in Louvain method, included in metadata.
        max_power (int, optional): Maximum power used in WGCNA, included in metadata.
        precluster (str, optional): Preclustering strategy used, included in metadata.
        merge (bool, optional): Whether merging of biclusters was performed, included in metadata.

    Returns:
        None: This function does not return any value, it writes the results to a file.
    """
    bics = biclusters.copy()
    if add_metadata:
        metadata = (
            "#seed="
            + str(seed)
            + "; "
            + "pval="
            + str(pval)
            + "; "
            + "min_n_samples="
            + str(min_n_samples)
            + "; "
        )
        metadata = metadata + "b=" + bin_method + "; "
        metadata = metadata + "c=" + clust_method + "; "
        if len(directions):
            metadata = metadata + "directions=" + "-".join(directions) + "; "
        if clust_method == "Louvain":
            metadata = (
                metadata
                + "similarity_cutoff="
                + str(similarity_cutoff)
                + "; modularity="
                + str(m)
            )
        elif "WGCNA" in clust_method:
            metadata = (
                metadata
                + "ds="
                + str(ds)
                + "; dch="
                + str(dch)
                + "; max_power="
                + str(max_power)
                + "; precluster="
                + str(precluster)
            )

        else:
            logger.error(f"Unknown 'clust_method': {clust_method}")
        metadata = metadata + "; merge=" + str(merge)
        with open(results_file_name, "w") as f:
            f.write(metadata + "\n")
        write_mode = "a"
    else:
        write_mode = "w"

    if len(bics) == 0:
        logger.warning("No biclusters found")
    else:
        if to_str:
            bics["genes"] = bics["genes"].apply(lambda x: " ".join(map(str, sorted(x))))
            bics["genes_up"] = bics["genes_up"].apply(
                lambda x: " ".join(map(str, sorted(x)))
            )
            bics["genes_down"] = bics["genes_down"].apply(
                lambda x: " ".join(map(str, sorted(x)))
            )
            bics["samples"] = bics["samples"].apply(
                lambda x: " ".join(map(str, sorted(x)))
            )
            if "gene_indexes" in bics.columns:
                bics["gene_indexes"] = bics["gene_indexes"].apply(
                    lambda x: " ".join(map(str, sorted(x)))
                )
            if "sample_indexes" in bics.columns:
                bics["sample_indexes"] = bics["sample_indexes"].apply(
                    lambda x: " ".join(map(str, sorted(x)))
                )
            if "ids" in bics.columns:
                bics["ids"] = bics["ids"].apply(lambda x: " ".join(map(str, sorted(x))))
        bics.index.name = "id"
    bics.to_csv(results_file_name, sep="\t", mode=write_mode)


class ProjectPaths:
    """
    A class to manage project paths for input and output files.
    """

    def _init_roots(self, out_dir: str, binary_dir: str) -> tuple[Path, Path, bool]:
        """
        Initialize root and bin_root directories

        Args:
            out_dir (str, optional): The base directory for runs. Defaults to "runs/run_<timestamp>".
            binary_dir (str, optional): The directory for binarization results. Defaults to "<out_dir>/binarization".

        Returns:
            Path: The calculated root directory path.
            Path: The calculated binary directory path.
            created = False if the directory already exists, True otherwise.
        """
        now_str = f"{datetime.now():%Y%m%dT%H%M%S}"
        out_dir = out_dir.replace("<timestamp>", now_str)
        binary_dir = binary_dir.replace("<timestamp>", now_str)
        binary_dir = binary_dir.replace("<out_dir>", out_dir)

        root = Path(out_dir).resolve()
        bin_dir = Path(binary_dir).resolve()

        # Create output directory if it doesn't exist
        created = False
        if not root.exists():
            root.mkdir(parents=True, exist_ok=True)
            created = True

        return root, bin_dir, created

    def __init__(
        self,
        out_dir: str = "runs/run_<timestamp>",
        binary_dir: str = "<out_dir>/binarization",
    ) -> None:
        """
        Initializes the ProjectPaths with the given directories.

        Args:
            out_dir (str): Output directory path.
            binary_dir (str): Directory to save/load binarization results.
        """
        # replace placeholders
        root, bin_dir, created = self._init_roots(out_dir, binary_dir)
        self.root = root
        self.bin_dir = bin_dir
        using = "Created new" if created else "Using existing"
        logger.debug(f"{using} directory for outputs: {self.root}")

        # Main output files
        self.args = str(self.root / "args.tsv")
        self.res = str(self.root / "biclusters.tsv")
        self.log = str(self.root / "unpast.log")

        # Set default binary_dir if not provided
        self.binarization_args = str(self.bin_dir / "bin_args.tsv")
        self.binarization_res = str(self.bin_dir / "bin_res.tsv")
        self.binarization_stats = str(self.bin_dir / "bin_stats.tsv")
        self.binarization_bg = str(self.bin_dir / "bin_background.tsv")

        self.tmp_wgcna = self.root / "tmp_wgcna"

    def create_binarization_paths(self) -> None:
        """
        Creates the binarization directory and ensures all binarization paths exist.
        This method is called before saving binarization files.
        """
        Path(self.bin_dir).mkdir(parents=True, exist_ok=True)

    def get_root_dir(self) -> str:
        """
        Returns the root directory of outputs.
        """
        return str(self.root)

    def get_wgcna_tmp_file(self) -> str:
        """
        Returns a temporary file path for WGCNA results.
        The file name includes a timestamp to ensure uniqueness.
        """
        self.tmp_wgcna.mkdir(exist_ok=True)
        self.tmp_wgcna_file = str(
            self.tmp_wgcna / f"wgcna_{datetime.now():%Y%m%dT%H%M%S}.tsv"
        )
        return self.tmp_wgcna_file

    def clear_wgcna_tmp_files(self) -> None:
        """
        Clears the temporary WGCNA file and folder by removing it if it exists.
        """
        try:
            os.remove(self.tmp_wgcna_file)
        except Exception as e:
            logger.warning(
                f"Failed to remove temporary WGCNA file: {self.tmp_wgcna_file}."
            )
            logger.warning(f"  Error: {e}")

        try:
            self.tmp_wgcna.rmdir()
        except Exception as e:
            logger.warning(
                f"Failed to remove temporary WGCNA directory: {self.tmp_wgcna}"
            )
            logger.warning(f"  Error: {e}")

    def __str__(self) -> str:
        """
        Returns a string representation of the ProjectPaths object.
        """
        return f"ProjectPaths(root='{self.root}')"


def write_args(
    args: dict, file_path: str, args_label: Optional[str] = "Arguments"
) -> None:
    """
    Writes the arguments to a file in a tab-separated format.

    Args:
        args (dict): A dictionary of arguments to save.
        file_path (str): The path to the file where the arguments will be saved.
        args_label (str, optional): The name of args, use None to skip logging.
    """
    assert file_path.endswith(".tsv"), "File for args must end with '.tsv'. "
    df = pd.DataFrame(list(args.items()), columns=["arg", "value"])
    df.to_csv(file_path, sep="\t", index=False)
    if args_label:
        logger.debug(f"{args_label} saved to {file_path}")


def read_args(file_path: str) -> dict:
    """
    Reads arguments from a file in a tab-separated format.

    Args:
        file_path (str): The path to the file from which to load the arguments.

    Returns:
        dict: A dictionary of loaded arguments.
    """
    assert file_path.endswith(".tsv"), "File for args must end with '.tsv'. "
    df = pd.read_csv(file_path, sep="\t")
    args = dict(zip(df["arg"], df["value"]))
    return args



"""Tests for preprocessing module."""

import numpy as np
import pandas as pd
import pytest

from unpast.core.preprocessing import prepare_input_matrix, zscore
from unpast.utils.logs import LOG_LEVELS


class TestZscore:
    """Test cases for zscore function."""

    def test_zscore_basic(self):
        """Test basic z-score normalization."""
        # Create a simple test matrix
        data = {"sample1": [1, 4, 7], "sample2": [2, 5, 8], "sample3": [3, 6, 9]}
        df = pd.DataFrame(data, index=["gene1", "gene2", "gene3"])

        result = zscore(df)

        # Check that means are approximately 0
        means = result.mean(axis=1)
        assert np.allclose(means, 0, atol=1e-10)

        # Check that standard deviations are approximately 1
        stds = result.std(axis=1)
        assert np.allclose(stds, 1, atol=1e-10)

    def test_zscore_zero_variance(self, caplog):
        """Test z-score normalization with zero variance genes."""
        # Create matrix with one zero-variance gene
        data = {"sample1": [1, 5, 5], "sample2": [2, 5, 5], "sample3": [3, 5, 5]}
        df = pd.DataFrame(data, index=["gene1", "gene2", "gene3"])

        # Capture warning messages using caplog
        with caplog.at_level(LOG_LEVELS["WARNING"]):
            result = zscore(df)

        # Check that zero variance genes are set to 0
        assert np.allclose(result.loc["gene2"], 0)
        assert np.allclose(result.loc["gene3"], 0)

        # Check that warning was logged
        assert "zero variance rows detected" in caplog.text
        assert "2" in caplog.text  # Should detect 2 zero variance rows

    def test_zscore_single_gene(self):
        """Test z-score normalization with single gene."""
        data = {"sample1": [1], "sample2": [2], "sample3": [3]}
        df = pd.DataFrame(data, index=["gene1"])

        result = zscore(df)

        # Check normalization worked
        assert np.allclose(result.mean(axis=1), 0, atol=1e-10)
        assert np.allclose(result.std(axis=1), 1, atol=1e-10)

    def test_zscore_preserves_shape(self):
        """Test that zscore preserves DataFrame shape and indices."""
        data = {
            "sample1": [1, 4, 7, 10],
            "sample2": [2, 5, 8, 11],
            "sample3": [3, 6, 9, 12],
        }
        df = pd.DataFrame(data, index=["gene1", "gene2", "gene3", "gene4"])

        result = zscore(df)

        # Check shape is preserved
        assert result.shape == df.shape

        # Check indices are preserved
        assert list(result.index) == list(df.index)
        assert list(result.columns) == list(df.columns)


class TestPrepareInputMatrix:
    """Test cases for prepare_input_matrix function."""

    def setup_method(self):
        """Set up test data for each test method."""
        # Create a test matrix with known properties
        np.random.seed(42)
        self.test_data = pd.DataFrame(
            np.random.randn(5, 10),
            index=["gene1", "gene2", "gene3", "gene4", "gene5"],
            columns=[f"sample{i}" for i in range(1, 11)],
        )

        # Create data that's already standardized
        self.standardized_data = pd.DataFrame(
            np.random.randn(3, 5),
            index=["gene1", "gene2", "gene3"],
            columns=["sample1", "sample2", "sample3", "sample4", "sample5"],
        )
        # Manually standardize it
        self.standardized_data = (
            self.standardized_data.T - self.standardized_data.mean(axis=1)
        ).T
        self.standardized_data = (
            self.standardized_data.T / self.standardized_data.std(axis=1)
        ).T

    def test_prepare_input_matrix_basic(self):
        """Test basic functionality of prepare_input_matrix."""
        result = prepare_input_matrix(self.test_data)

        # Check that result is standardized
        means = result.mean(axis=1)
        stds = result.std(axis=1)
        assert np.allclose(means, 0, atol=1e-10)
        assert np.allclose(stds, 1, atol=1e-10)

        # Check that indices are strings
        assert all(isinstance(idx, str) for idx in result.index)
        assert all(isinstance(col, str) for col in result.columns)

    def test_prepare_input_matrix_already_standardized(self, caplog):
        """Test with already standardized data."""
        with caplog.at_level(LOG_LEVELS["INFO"]):
            result = prepare_input_matrix(self.standardized_data)

        # Should not print standardization message
        assert "Input is not standardized" not in caplog.text

        # Result should be approximately the same as input
        assert np.allclose(result.values, self.standardized_data.values, atol=1e-10)

    def test_prepare_input_matrix_zero_variance(self, caplog):
        """Test handling of zero variance genes."""
        # Add zero variance genes
        data_with_zeros = self.test_data.copy()
        data_with_zeros.loc["zero_gene1"] = 5.0  # Constant value
        data_with_zeros.loc["zero_gene2"] = 10.0  # Another constant value

        with caplog.at_level(LOG_LEVELS["DEBUG"]):
            result = prepare_input_matrix(data_with_zeros)

        # Check that zero variance genes were dropped
        assert "zero_gene1" not in result.index
        assert "zero_gene2" not in result.index
        assert result.shape[0] == self.test_data.shape[0]  # Original number of genes

        # Check verbose output
        assert "Zero variance rows will be dropped: 2" in caplog.text

    def test_prepare_input_matrix_with_missing_values(self, caplog):
        """Test handling of missing values."""
        # Add missing values
        data_with_na = self.test_data.copy()
        data_with_na.iloc[0, :3] = np.nan  # gene1 has 3 missing values
        data_with_na.iloc[1, :8] = (
            np.nan
        )  # gene2 has 8 missing values (should be dropped)

        with caplog.at_level(LOG_LEVELS["WARNING"]):
            result = prepare_input_matrix(data_with_na, min_n_samples=5)

        # gene2 should be dropped (only 2 valid samples, less than min_n_samples=5)
        assert "gene2" not in result.index
        assert "gene1" in result.index  # gene1 should be kept (7 valid samples)

        # Check verbose output
        assert "Missing values detected" in caplog.text
        assert "Features with too few values" in caplog.text

    def test_prepare_input_matrix_with_ceiling(self, caplog):
        """Test z-score ceiling functionality."""
        # Create data with extreme values
        data = pd.DataFrame(
            [[10, 1, 1], [1, 10, 1], [1, 1, 10]],
            index=["gene1", "gene2", "gene3"],
            columns=["sample1", "sample2", "sample3"],
        )

        with caplog.at_level(LOG_LEVELS["DEBUG"]):
            result = prepare_input_matrix(data, ceiling=2.0)

        # Check that values are capped at ceiling
        assert np.all(result <= 2.0)
        assert np.all(result >= -2.0)

        # Check verbose output
        assert "Standardized expressions will be limited to [-2.0,2.0]:" in caplog.text

    def test_prepare_input_matrix_no_standardization(self):
        """Test with standardization disabled."""
        result = prepare_input_matrix(self.test_data, standradize=False)

        # Result should be the same as input (except for string conversion)
        expected = self.test_data.copy()
        expected.index = expected.index.astype(str)
        expected.columns = expected.columns.astype(str)

        pd.testing.assert_frame_equal(result, expected)

    def test_prepare_input_matrix_missing_values_with_ceiling(self, caplog):
        """Test missing values replacement with ceiling."""
        # Create data with missing values
        data = self.test_data.copy()
        data.iloc[0, 0] = np.nan
        data.iloc[1, 1] = np.nan

        with caplog.at_level(LOG_LEVELS["DEBUG"]):
            result = prepare_input_matrix(data, ceiling=2.0)

        # Check that missing values were replaced with -ceiling
        # Note: After standardization, the exact values will be different,
        # but we can check that no NaN values remain
        assert not result.isna().any().any()

        # Check verbose output
        assert "Missing values will be replaced with -2.0." in caplog.text

    def test_prepare_input_matrix_too_few_features(self, caplog):
        """Test behavior when too few features remain after filtering."""
        # Create data where most genes will be dropped
        data = pd.DataFrame(
            [
                [1, 1, 1],
                [2, 2, 2],
            ],  # Only 2 genes, both will have zero variance after standardization
            index=["gene1", "gene2"],
            columns=["sample1", "sample2", "sample3"],
        )

        with caplog.at_level(LOG_LEVELS["WARNING"]):
            # This should trigger the warning about too few features
            _ = prepare_input_matrix(data)

        # The exact behavior may vary, but we should get a warning
        assert "less than 3 features (rows) remain" in caplog.text

    def test_prepare_input_matrix_duplicate_indices(self, caplog):
        """Test handling of non-unique row names."""
        # Create data with duplicate row names
        data = self.test_data.copy()
        data = data.reindex(["gene1", "gene1", "gene2", "gene2", "gene3"])  # Duplicates

        with caplog.at_level(LOG_LEVELS["WARNING"]):
            _ = prepare_input_matrix(data)

        # Should print warning about non-unique row names
        assert "Row names are not unique" in caplog.text

    def test_prepare_input_matrix_tolerance(self):
        """Test the tolerance parameter for standardization check."""
        # Create data that's almost standardized
        data = self.standardized_data.copy()
        data = data + 0.005  # Add small deviation

        # With default tolerance (0.01), should not re-standardize
        result1 = prepare_input_matrix(data, tol=0.01)

        # With smaller tolerance (0.001), should re-standardize
        result2 = prepare_input_matrix(data, tol=0.001)

        # Results should be different
        assert not np.allclose(result1.values, result2.values)

    def test_prepare_input_matrix_min_n_samples(self):
        """Test the min_n_samples parameter."""
        # Create data with missing values
        data = pd.DataFrame(
            np.random.randn(3, 10),
            index=["gene1", "gene2", "gene3"],
            columns=[f"sample{i}" for i in range(1, 11)],
        )
        # Add missing values to gene1 (6 missing, 4 valid)
        data.iloc[0, :6] = np.nan

        # With min_n_samples=5, gene1 should be dropped
        result1 = prepare_input_matrix(data, min_n_samples=5)
        assert "gene1" not in result1.index

        # With min_n_samples=3, gene1 should be kept
        result2 = prepare_input_matrix(data, min_n_samples=3)
        assert "gene1" in result2.index


if __name__ == "__main__":
    pytest.main([__file__])



# Usage: from unpast.misc.unpast_DE import run_de_for_unpast
# run_de_for_unpast(unpast_output_path, expression_matrix_path, counts = False, [keep_all=False,adj_p_value_cut_off = 0.05, logFC_cut_off = 1, r_script_path = None, r_executable_path = None])

import logging
import os
import subprocess
from pathlib import Path

import numpy as np
import pandas as pd

# Add all logging levels
logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")

####################################################################################################
# Constants
####################################################################################################
DELIMITER = "\t"
RSCRIPTS_DIR = Path(__file__).resolve().parent.parent / "rscripts"

####################################################################################################
# Functions
####################################################################################################


def extract_samples_to_file(unpast_df: pd.DataFrame, samples_to_compare: str) -> None:
    """
    Extract samples from the unpast dataframe.
    """
    samples = unpast_df[["samples", "n_samples"]]
    samples.to_csv(samples_to_compare, sep=DELIMITER, index_label="")


def run_add_genes_script(
    samples_to_compare: str,
    expression_matrix_path: str,
    counts: bool = False,
    adj_p_value_cut_off: float = 0.05,
    logFC_cut_off: float = 1,
    num_genes_cut_off=float("inf"),
    r_script_path: str = None,
    r_executable_path: str = None,
) -> str:
    if not r_script_path:
        r_script_path = str(RSCRIPTS_DIR / "add_genes.R")

    if r_executable_path is None:
        r_executable_path = "Rscript"
    else:
        r_executable_path = os.path.join(r_executable_path, "Rscript")

    logging.debug(
        [
            r_executable_path,
            r_script_path,
            samples_to_compare,
            expression_matrix_path,
            str(counts).upper(),
            str(adj_p_value_cut_off),
            str(logFC_cut_off),
            str(num_genes_cut_off),
        ]
    )

    process = subprocess.Popen(
        [
            r_executable_path,
            r_script_path,
            samples_to_compare,
            expression_matrix_path,
            str(counts).upper(),
            str(adj_p_value_cut_off),
            str(logFC_cut_off),
            str(num_genes_cut_off),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()

    # Decode the output and error messages
    samples_with_genes_path = stdout.decode("utf-8")

    if stdout:
        logging.debug(
            "DE analysis temporary output was generated: %s", samples_with_genes_path
        )
    if stderr:
        error_message = stderr.decode("utf-8")
        logging.error("Error Output: %s", error_message)
        raise Exception(f"Error in subprocess: {error_message}")

    return samples_with_genes_path


def filter_de_genes(
    new_unpast_df: pd.DataFrame, columns_to_process: list
) -> pd.DataFrame:
    """
    Filter de_genes_df to keep only the genes that are in the unpast_df.
    """
    # Loop through each column
    for col in columns_to_process:
        col_DE = col + "_DE"

        # Check for NaN in original column and replace the corresponding _DE column value with NaN
        new_unpast_df[col_DE] = np.where(
            new_unpast_df[col].isna(), np.nan, new_unpast_df[col_DE]
        )

        # Split the string into a list for both columns
        new_unpast_df[col] = new_unpast_df[col].apply(
            lambda x: str(x).split() if pd.notna(x) else []
        )
        new_unpast_df[col_DE] = new_unpast_df[col_DE].apply(
            lambda x: str(x).split() if pd.notna(x) else []
        )

        # Keep in the _DE column only values that are in the corresponding original column
        new_unpast_df[col_DE] = new_unpast_df.apply(
            lambda x: [i for i in x[col_DE] if i in x[col]], axis=1
        )

        # Join the list back into a string
        new_unpast_df[col] = new_unpast_df[col].apply(
            lambda x: " ".join(x) if x else np.nan
        )
        new_unpast_df[col_DE] = new_unpast_df[col_DE].apply(
            lambda x: " ".join(x) if x else np.nan
        )
    return new_unpast_df


def add_columns_to_unpast_df(
    unpast_df: pd.DataFrame, de_genes_df: pd.DataFrame, keep_all: bool = False
) -> pd.DataFrame:
    """
    Filter de_genes_df to keep only the genes that are in the unpast_df.
    Add four new columns to the original unpast table.
    """
    # add columns to unpast_df
    new_unpast_df = unpast_df.assign(
        n_genes_DE=de_genes_df["n_genes"],
        genes_DE=de_genes_df["genes"],
        genes_up_DE=de_genes_df["genes_up"],
        genes_down_DE=de_genes_df["genes_down"],
    )

    # filter de_genes_df to keep only the genes that are in the unpast_df
    if not keep_all:
        logging.info("Filtering DE genes not in the original UnPaSt output")
        new_unpast_df = filter_de_genes(
            new_unpast_df, columns_to_process=["genes", "genes_down", "genes_up"]
        )

    # update n_genes_DE column with the number of genes in the genes_DE column
    new_unpast_df["n_genes_DE"] = new_unpast_df["genes_DE"].apply(
        lambda x: len(x.split()) if pd.notna(x) else 0
    )

    return new_unpast_df


def read_dataframe_from_file(file_path: str) -> pd.DataFrame:
    if not os.path.isfile(file_path):
        logging.error("File does not exist: %s", file_path)
        raise FileNotFoundError(f"File does not exist: {file_path}")

    if os.stat(file_path).st_size == 0:
        logging.error("File is empty: %s", file_path)
        raise ValueError(f"File is empty: {file_path}")

    df = pd.read_csv(file_path, delimiter=DELIMITER, comment="#", index_col=0)

    if df.empty:
        logging.error("UnPaSt output is empty: %s", file_path)
        raise ValueError(f"Dataframe is empty: {file_path}")

    return df


def write_result(
    df: pd.DataFrame, input_file_path: str, output_file_path: str
) -> pd.DataFrame:
    # Checking if input file exists and if it's not empty
    if not os.path.isfile(input_file_path):
        logging.error("Input file does not exist: %s", input_file_path)
        raise FileNotFoundError(f"Input file does not exist: {input_file_path}")

    if os.stat(input_file_path).st_size == 0:
        logging.error("Input file is empty: %s", input_file_path)
        raise ValueError(f"Input file is empty: {input_file_path}")

    # Check if output directory is writable
    output_dir_name = os.path.dirname(output_file_path)
    if not os.access(output_dir_name, os.W_OK):
        logging.error("Output directory is not writable: %s", output_dir_name)
        raise PermissionError(f"Output directory is not writable: {output_dir_name}")

    with open(input_file_path, "r") as f_in, open(output_file_path, "w") as f_out:
        comment_line = f_in.readline()
        f_out.write(comment_line)
        df.to_csv(f_out, sep=DELIMITER)


def safe_remove(file_path: str) -> None:
    try:
        os.remove(file_path)
    except OSError as e:
        logging.error(f"Error: {file_path} : {e.strerror}")


def run_de_for_unpast(
    unpast_output_path: str,
    expression_matrix_path: str,
    counts: bool = False,
    keep_all: bool = False,
    adj_p_value_cut_off: float = 0.05,
    logFC_cut_off: float = 1,
    num_genes_cut_off: float = float("inf"),
    r_script_path: str = None,
    r_executable_path: str = None,
) -> None:
    """
    A function that runs differential expression analysis using the limma package in R for unpast biclusters.
    """
    # Read unpast output
    # Read file with rownames in the first columns, first line is the comment line, colnames in the second line
    logging.info("Reading UnPaSt biclusters output and extracting samples to compare")
    # Check file existance, emptiness, and read DataFrame
    unpast_df = read_dataframe_from_file(unpast_output_path)
    # get the samples to compare
    samples_to_compare = os.path.join(
        os.path.dirname(unpast_output_path),
        f"{os.path.splitext(os.path.basename(unpast_output_path))[0][: -len('_biclusters')]}_samples{os.path.splitext(os.path.basename(unpast_output_path))[1]}",
    )
    # save samples to file
    extract_samples_to_file(unpast_df, samples_to_compare)

    # run add_genes.R script
    logging.info("Running DE analysis for unpast biclusters, takes a while...")
    samples_with_genes_path = run_add_genes_script(
        samples_to_compare,
        expression_matrix_path,
        counts,
        adj_p_value_cut_off,
        logFC_cut_off,
        num_genes_cut_off,
        r_script_path,
        r_executable_path,
    )

    logging.info("Adding DE genes to UnPaSt output table")
    # Read the unpast output file into a pandas dataframe.
    # Read file with rownames in the first columns, first line is the comment line, colnames in the second line
    de_genes_df = pd.read_csv(
        samples_with_genes_path.strip(), delimiter=DELIMITER, header=0, index_col=0
    )

    # add columns to unpast_df and filter de_genes_df to keep only the genes that are in the unpast_df
    new_unpast_df = add_columns_to_unpast_df(unpast_df, de_genes_df, keep_all=keep_all)

    # write new_unpast_df to file
    # use the original unpast_df file name and add _DE to the end
    # also add the comment line from the original unpast_df to the top of the new file
    output_path_de = unpast_output_path.replace(
        ".tsv", f"_DE.pval{adj_p_value_cut_off}.logFC{logFC_cut_off}.tsv"
    )
    # write_result(new_unpast_df, unpast_output_path, output_path_de)

    # remove the temporary files
    logging.info("Removing temporary files")
    safe_remove(samples_to_compare)
    safe_remove(samples_with_genes_path.strip())

    # keep only columns with genes
    cols = [
        "n_genes",
        "genes",
        "n_genes_DE",
        "genes_DE",
        "genes_up_DE",
        "genes_down_DE",
    ]
    de_df = new_unpast_df.loc[:, cols]
    cols = ["genes", "genes_DE", "genes_up_DE", "genes_down_DE"]
    de_df.loc[:, cols] = (
        de_df.loc[:, cols].fillna("").applymap(lambda row: set(row.split(" ")))
    )
    return de_df
